package dev.fitko.fitconnect.core.events;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.JWSAlgorithm;
import com.nimbusds.jose.JWSVerifier;
import com.nimbusds.jose.crypto.RSASSAVerifier;
import com.nimbusds.jose.jwk.JWK;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jwt.SignedJWT;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.Environment;
import dev.fitko.fitconnect.api.config.EnvironmentName;
import dev.fitko.fitconnect.api.config.defaults.SchemaConfig;
import dev.fitko.fitconnect.api.domain.model.event.Event;
import dev.fitko.fitconnect.api.domain.model.event.EventPayload;
import dev.fitko.fitconnect.api.domain.model.event.problems.submission.AttachmentsMismatch;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.schema.SchemaResources;
import dev.fitko.fitconnect.api.exceptions.internal.EventCreationException;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.api.services.events.SecurityEventService;
import dev.fitko.fitconnect.api.services.schema.SchemaProvider;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import dev.fitko.fitconnect.core.crypto.HashService;
import dev.fitko.fitconnect.core.crypto.JWECryptoService;
import dev.fitko.fitconnect.core.http.HttpClient;
import dev.fitko.fitconnect.core.schema.SchemaResourceProvider;
import dev.fitko.fitconnect.core.util.CertificateLoader;
import dev.fitko.fitconnect.core.validation.DefaultValidationService;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.text.ParseException;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.Mockito.mock;

class SecurityEventTokenServiceTest {

    private RSAKey signingKey;
    private RSAKey encryptionKey;
    private CryptoService cryptoService;
    private ValidationService validationService;
    private SecurityEventService underTest;
    private static Locale defaultLocale;

    @BeforeEach
    void setUp() throws IOException, ParseException {

        defaultLocale = Locale.getDefault();
        Locale.setDefault(Locale.ENGLISH);

        final ApplicationConfig config = getApplicationConfig();
        cryptoService = new JWECryptoService(new HashService());
        signingKey = JWK.parse(getResourceAsString("private_test_signing_key.json")).toRSAKey();
        encryptionKey = JWK.parse(getResourceAsString("public_encryption_test_key.json")).toRSAKey();

        final List<String> setSchemas = SchemaConfig.getSetSchemaFilePaths("/set-schema");
        final List<String> metadataSchemas = SchemaConfig.getMetadataSchemaFileNames("/metadata-schema");
        final List<String> destinationSchemas = SchemaConfig.getDestinationSchemaPaths("/destination-schema");
        final SchemaResources schemaResources = new SchemaResources(setSchemas, metadataSchemas, destinationSchemas, config.getSubmissionDataSchemas());
        final SchemaProvider schemaProvider = new SchemaResourceProvider(mock(HttpClient.class), schemaResources);

        validationService = new DefaultValidationService(config, new HashService(), schemaProvider,
                CertificateLoader.create().loadTrustedRootCertificates(List.of("/trusted-test-root-certificates/TEST-PCA20.pem")));
        underTest = new SecurityEventTokenService(config, validationService, signingKey);
    }
    @AfterAll
    static void tearDown() {
        Locale.setDefault(defaultLocale);
    }

    @Test
    void testAcceptSubmissionEventWithoutProblems() throws JOSEException {

        // Given
        final var encryptedData = cryptoService.encryptObject(encryptionKey, "test data", "application/json");
        final var encryptedMetadata = cryptoService.encryptObject(encryptionKey, "test metadata", "application/json");

        final var encryptedAttachments = new HashMap<UUID, String>();
        encryptedAttachments.put(UUID.randomUUID(), cryptoService.encryptObject(encryptionKey, "test attachment 1", "application/json"));
        encryptedAttachments.put(UUID.randomUUID(), cryptoService.encryptObject(encryptionKey, "test attachment 2", "application/json"));

        final var submission = new Submission();
        submission.setSubmissionId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());
        submission.setCaseId(UUID.randomUUID());
        submission.setEncryptedData(encryptedData);
        submission.setEncryptedMetadata(encryptedMetadata);

        final EventPayload eventPayload = EventPayload.forAcceptEventWithAttachments(submission, encryptedAttachments);

        // When
        final SignedJWT signedJWT = underTest.createAcceptSubmissionEvent(eventPayload);

        // Then
        final Map<String, Object> payload = signedJWT.getPayload().toJSONObject();
        final var eventsClaim = (Map) payload.get("events");
        final var eventObject = (Map) eventsClaim.get(Event.ACCEPT.getSchemaUri());
        final JWSVerifier jwsVerifier = new RSASSAVerifier(signingKey);

        assertEquals(JWSAlgorithm.PS512, signingKey.getAlgorithm());
        assertTrue(signedJWT.verify(jwsVerifier));
        assertThat(payload.get("sub"), is("submission:" + eventPayload.getSubmissionId()));
        assertThat(payload.get("txn"), is("case:" + eventPayload.getCaseId()));
        assertThat(payload.get("iss"), is(eventPayload.getDestinationId().toString()));
        assertThat(eventObject.size(), is(1));
    }

    @Test
    void testAcceptSubmissionEventWithProblem() throws JOSEException {

        // Given
        final var encryptedData = cryptoService.encryptObject(encryptionKey, "test data", "application/json");
        final var encryptedMetadata = cryptoService.encryptObject(encryptionKey, "test metadata", "application/json");

        final var encryptedAttachments = new HashMap<UUID, String>();
        encryptedAttachments.put(UUID.randomUUID(), cryptoService.encryptObject(encryptionKey, "test attachment", "application/json"));

        final var submission = new Submission();
        submission.setSubmissionId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());
        submission.setCaseId(UUID.randomUUID());
        submission.setEncryptedData(encryptedData);
        submission.setEncryptedMetadata(encryptedMetadata);

        final var problem = new AttachmentsMismatch();

        final EventPayload eventPayload = EventPayload.forAcceptEventWithAttachments(submission, encryptedAttachments, problem);

        // When
        final SignedJWT signedJWT = underTest.createAcceptSubmissionEvent(eventPayload);

        // Then
        final Map<String, Object> payload = signedJWT.getPayload().toJSONObject();
        final var eventsClaim = (Map) payload.get("events");
        final var eventObject = (Map) eventsClaim.get(Event.ACCEPT.getSchemaUri());
        final JWSVerifier jwsVerifier = new RSASSAVerifier(signingKey);

        assertEquals(JWSAlgorithm.PS512, signingKey.getAlgorithm());
        assertTrue(signedJWT.verify(jwsVerifier));
        assertThat(payload.get("sub"), is("submission:" + eventPayload.getSubmissionId()));
        assertThat(payload.get("txn"), is("case:" + eventPayload.getCaseId()));
        assertThat(payload.get("iss"), is(eventPayload.getDestinationId().toString()));
        assertThat(eventObject.size(), is(2));
        assertThat(((List) eventObject.get("problems")).get(0), is(problem));
    }

    @Test
    void testRejectSubmissionEvent() throws JOSEException {

        // Given
        final var problem = new AttachmentsMismatch();

        final var submission = new Submission();
        submission.setSubmissionId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());
        submission.setCaseId(UUID.randomUUID());

        final EventPayload eventPayload = EventPayload.forRejectEvent(submission, List.of(problem));

        // When
        final SignedJWT signedJWT = underTest.createRejectSubmissionEvent(eventPayload);

        // Then
        final Map<String, Object> payload = signedJWT.getPayload().toJSONObject();
        final var eventsClaim = (Map) payload.get("events");
        final var eventObject = (Map) eventsClaim.get(Event.REJECT.getSchemaUri());
        final JWSVerifier jwsVerifier = new RSASSAVerifier(signingKey);

        assertEquals(JWSAlgorithm.PS512, signingKey.getAlgorithm());
        assertTrue(signedJWT.verify(jwsVerifier));
        assertThat(payload.get("sub"), is("submission:" + eventPayload.getSubmissionId()));
        assertThat(payload.get("txn"), is("case:" + eventPayload.getCaseId()));
        assertThat(payload.get("iss"), is(eventPayload.getDestinationId().toString()));
        assertThat(eventObject.size(), is(1));
        assertThat(((List) eventObject.get("problems")).get(0), is(problem));
    }

    @Test
    void testAcceptEventWithMissingMetadataAuthTag() {

        // Given
        final var submission = new Submission();
        submission.setSubmissionId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());
        submission.setCaseId(UUID.randomUUID());
        submission.setEncryptedData(cryptoService.encryptObject(encryptionKey, "test data", "application/json"));

        final EventPayload eventPayload = EventPayload.forAcceptEvent(submission);

        // When
        final EventCreationException exception = assertThrows(EventCreationException.class, () -> underTest.createAcceptSubmissionEvent(eventPayload));

        // Then
        assertThat(exception.getCause().getMessage(), containsString("$.events.https://schema.fitko.de/fit-connect/events/accept-submission.authenticationTags.metadata: null found"));
    }

    @Test
    void testAcceptEventWithMissingDataAuthTag() {

        // Given
        final var submission = new Submission();
        submission.setSubmissionId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());
        submission.setCaseId(UUID.randomUUID());
        submission.setEncryptedMetadata(cryptoService.encryptObject(encryptionKey, "test metadata", "application/json"));

        final EventPayload eventPayload = EventPayload.forAcceptEvent(submission);

        // When
        final EventCreationException exception = assertThrows(EventCreationException.class, () -> underTest.createAcceptSubmissionEvent(eventPayload));

        // Then
        assertThat(exception.getCause().getMessage(), containsString("$.events.https://schema.fitko.de/fit-connect/events/accept-submission.authenticationTags.data: null found"));
    }

    @Test
    void testAcceptEventWithMissingAuthData() {

        // Given
        final var submission = new Submission();
        submission.setSubmissionId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());
        submission.setCaseId(UUID.randomUUID());

        final EventPayload eventPayload = EventPayload.forAcceptEvent(submission);

        // When
        final EventCreationException exception = assertThrows(EventCreationException.class, () -> underTest.createAcceptSubmissionEvent(eventPayload));

        // Then
        assertThat(exception.getCause().getMessage(), containsString("$.events.https://schema.fitko.de/fit-connect/events/accept-submission.authenticationTags.metadata: null found"));
        assertThat(exception.getCause().getMessage(), containsString("$.events.https://schema.fitko.de/fit-connect/events/accept-submission.authenticationTags.data: null found"));
    }

    @Test
    void testRejectEventWithMissingProblems() {

        // Given
        final var submission = new Submission();
        submission.setSubmissionId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());
        submission.setCaseId(UUID.randomUUID());

        final EventPayload eventPayload = EventPayload.forAcceptEvent(submission);

        // When
        final EventCreationException exception = assertThrows(EventCreationException.class, () -> underTest.createRejectSubmissionEvent(eventPayload));

        // Then
        assertThat(exception.getCause().getMessage(), containsString("$.events.https://schema.fitko.de/fit-connect/events/reject-submission: required property 'problems' not found"));
    }

    private static ApplicationConfig getApplicationConfig() {
        final var envName = new EnvironmentName("testing");
        final var testing = new Environment();
        testing.setAllowInsecurePublicKey(true);
        return ApplicationConfig.builder()
                .activeEnvironment(envName)
                .environments(Map.of(envName, testing))
                .build();
    }

    private String getResourceAsString(final String name) throws IOException {
        final ClassLoader classLoader = getClass().getClassLoader();
        return Files.readString(new File(classLoader.getResource(name).getFile()).toPath());
    }
}